#!/usr/bin/env python
__author__      = "Olman Quiros Jimenez"
__copyright__   = "Copyright 2024"
__license__     = "MIT"

import socket
from typing import Dict

def recvall(sock: socket.socket) -> bytearray:
    '''
    Helper function to recv n bytes or return None if EOF is hit
    ''' 
    BUFF_SIZE = 1024
    data = bytearray()
    while True:
        part = sock.recv(BUFF_SIZE)
        data.extend(part)
        if len(part) < BUFF_SIZE:
            break
    return data





class BaseNode():
    def __init__(self, name:str, ip_addr:str='localhost', port:int=4200) -> 'BaseNode':
        self._ip_addr     = ip_addr
        self._port        = port
        super(BaseNode, self).__init__(name)

    def __repr__(self) -> str:
        return f"Node( '{self._name}', ip_addr='{self._ip_addr}', port={self._port}, items={self.items()})"


    def show(self):
        print(f"Node '{self._name}': (ip_addr='{self._ip_addr}', port={self._port})")
        if self.items():
            for k in self.items():
                print(f"\t{k}")
        else:
            print("\t...no items...")


    def load(self, data:Dict):
        for k, v in data.items():
            self[k] = v
            




#!/usr/bin/env python
__author__      = "Olman Quiros Jimenez"
__copyright__   = "Copyright 2024"
__license__     = "MIT"


from enum import Enum
from pickle import GET
 


class RequestCode(Enum):
    GET         = 1
    SET         = 2
    CLONE       = 3
    PULL        = 4
    SUBSCRIBE   = 5


class ResponseCode(Enum):
    OK                  = 1
    INVALID_REQUEST     = 2 
    INVALID_ARGS        = 3
    SERVER_ERROR        = 4



class RequestMessage:
    code:int = 0
    args:str = None


    def __init__(self, code:RequestCode, args:dict={}) -> 'RequestMessage':
        self.code = code.value
        self.args = args




class ResponseMessage:
    code:int = 0
    args:str = None

    def __init__(self, code:ResponseCode, args:dict={}) -> 'ResponseMessage':
        self.code = code.value
        self.args = args


#!/usr/bin/env python
__author__      = "Olman Quiros Jimenez"
__copyright__   = "Copyright 2024"
__license__     = "MIT"


class LibS3Error(Exception):
    """Base class for libs3 exceptions"""
    pass


#!/usr/bin/env python
__author__      = "Olman Quiros Jimenez"
__copyright__   = "Copyright 2024"
__license__     = "MIT"

import socket
import pickle
import threading
from typing import Dict, List, Any, Union

from s3py import LibS3Error, Param, Sys, Group, Param
from s3py.io import RequestMessage, ResponseMessage, RequestCode, ResponseCode
from .base_node import *





class Node(BaseNode):
    '''
    Node class.
    '''
    def __init__(self, name:str, ip_addr:str='localhost', port:int=4200, containers:List[Union[Param,Sys,Group]]=None) -> 'Node':
        super().__init__(name, ip_addr, port)
        for obj in containers:
            self.add(obj)

    def __setitem__(self, key, value) -> None:
        return self.add(value)

    def add(self, obj:Union[Param, Sys, Group]):
        '''
        Add a 'Param', 'Collection' or 'Group' to node.
        '''
        if type(obj) ==  Param or type(obj) == Sys or type(obj) == Group:
            return super().__setitem__(obj.name, obj)
        else:
            raise LibS3Error(f"Invalid type of object to add to Node '{self._name}'.  Object has to be 'Collection' or 'Group'.")



    def _handle_get(self, request:RequestMessage):
        try:
            param = self.query(request.args["param"])
            response_payload = {'value': param.get(),
                                'type': param.type
                                }
            return ResponseCode.OK, response_payload
        except:
            return ResponseCode.INVALID_REQUEST, {}


    def _handle_set(self, request:RequestMessage):
        try:
            param = self.query(request.args["param"])
            param.decode(request.args["value"])
            response_payload = {'value': param.get(),
                                'type': param.type
                                }
            return ResponseCode.OK, response_payload
        except Exception as e:
            response_code, response_payload = ResponseCode.INVALID_REQUEST, {}
            print(e)
            return ResponseCode.INVALID_REQUEST, {}


    def _handle_clone(self, request):
        try:
            response_payload = self.copy()
            return ResponseCode.OK, response_payload
        except Exception as e:
            response_code, response_payload = ResponseCode.INVALID_REQUEST, {}
            return ResponseCode.INVALID_REQUEST, {}



    def _handle_pull(self, request):
        pass



    def _handle_subscribe(self, request):
        pass



    def request_handler(self, cli_sock: socket.socket, cli_addr:List):
        try:
            while True:
                data =  recvall(cli_sock)
                request = pickle.loads(data)
                print(f"[*] From {cli_addr[0]}: {RequestCode(request.code).name} request, args: {request.args}")
                response_code    = None
                response_payload = None
                # Lookup Request handler
                if request.code == RequestCode.GET.value:
                    response_code, response_payload = self._handle_get(request)
                elif request.code == RequestCode.SET.value:
                    response_code, response_payload = self._handle_set(request)
                elif request.code == RequestCode.CLONE.value:
                    response_code, response_payload = self._handle_clone(request)
                elif request.code == RequestCode.PULL.value:
                    response_code, response_payload = self._handle_pull(request)
                elif request.code == RequestCode.SUBSCRIBE.value:
                    response_code, response_payload = self._handle_subscribe(request)
                else:
                    response_code, response_payload = ResponseCode.INVALID_REQUEST, {}
                # Write back to the client
                response = pickle.dumps(ResponseMessage(response_code, response_payload))
                cli_sock.sendall(response)
        except EOFError:
            print(f"[*] Client at {cli_addr[0]} disconnected.")
        finally:
            cli_sock.close()



    def serve_forever(self):
        '''
        Start Node as Server.
        '''   
        self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server.bind((self._ip_addr, self._port))
        self.server.listen(5)
        print(f"[*] Node '{self._name}' listening at {self._ip_addr}:{self._port}...")
        while True:
            cli_sock, cli_addr = self.server.accept()
            print(f"[*] Accepted connection from: {cli_addr[0]}:{cli_addr[1]}")
            client_handler = threading.Thread(target = self.request_handler, args=(cli_sock, cli_addr))
            client_handler.start()
    


class RemoteNode(BaseNode):
    '''
    RemoteNode to locally instantiate a Node 'living' in another virtual/physical context.
    '''
    def __init__(self, name:str, ip_addr:str='localhost', port:int=4200) -> 'RemoteNode':
        super().__init__(name, ip_addr, port)
        self.connect()

    def _transaction(self, request: RequestMessage) -> Dict:
        payload = pickle.dumps(request)     
        self.sock.send(payload)
        response = recvall(self.sock)
        if len(response):
            response = pickle.loads(response)
            print(f"From Server: {ResponseCode(response.code).name} response, args: {response.args}")
            return response
        else:
            return {}

    def __setitem__(self, key, value) :
        return super().__setitem__(key, value)


    def __getitem__(self, key):
        return super().__getitem__(key)

    def connect(self):
        self.sock: socket.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM) 
        self.sock.connect((self._ip_addr, self._port))

    def disconnect(self):
        self.sock.close()

    def get(self, param_path:str):
        response = self._transaction(RequestMessage(RequestCode.GET, 
                                                    { "param":param_path}))
        return response.args['value']

    def set(self, param_path:str, value:Any):
        response = self._transaction(RequestMessage(RequestCode.SET, 
                                                    { "param":param_path, 
                                                      "value":value}))
        


    def clone(self):
        response = self._transaction(RequestMessage(RequestCode.CLONE))
        self.load(response.args)

    def pull(self):
        args ={}
        for k, v in self.items():
            args[k] = v.get_commit()
        response = self._transaction(RequestMessage(RequestCode.PULL))

    def subscribe(self):
        response = self._transaction(RequestMessage(RequestCode.SUBSCRIBE))



from .error import LibS3Error
from .name import Name
from .path import Path, NamePath
from .sys import Sys, _root
from .param import *
from .group import Group, _group_list
from .param_db import ParamDB




#!/usr/bin/env python
__author__      = "Olman Quiros Jimenez"
__copyright__   = "Copyright 2024"
__license__     = "MIT"




def pprint_param_line(param, depth, index:bool, type:bool, size:bool) -> str:
     result = ""
     if index != False:
          result += "{:<5}".format(param.index)

     if type != False:
          result += "{:<5}".format(param.type)

     if size != False:
          result += "{:<5}".format(param.size)

     result += "| " * depth
     result += param.name
     result += ":\t{}".format(param.get())
     return result



def pprint_sys_line(obj, depth, index:bool, type:bool, size:bool) -> str:
     result = ""
     if index != False:
          result += "{:<5}".format(obj.index)

     if type != False:
          result += "{:<5}".format(obj.type)

     if size != False:
          result += "{:<5}".format(obj.size)

     result += "| " * depth
     result += obj.name
     return result


#!/usr/bin/env python
__author__      = "Olman Quiros Jimenez"
__copyright__   = "Copyright 2024"
__license__     = "MIT"


from typing import Any
import unittest
from datetime import datetime
from s3py import LibS3Error, Name, Sys, Path
from enum import Enum
import struct


class ParamType(Enum):
    BOOL   = "bool"
    UINT8  = "u8"
    INT8   = "i8"
    UINT16 = "u16"
    INT16  = "i16"
    UINT32 = "u32"
    INT32  = "i32"
    UINT64 = "u64"
    INT64  = "i64"
    FLT32  = "f32"
    FLT64  = "f64"
    DATE   = "date"
    STR    = "str"
    BLOB   = "blob"


    @property
    def encode_specifier(self):
        specifiers = {
            ParamType.BOOL.value   : "B",
            ParamType.UINT8.value  : "B",
            ParamType.INT8.value   : "b",
            ParamType.UINT16.value : "H",
            ParamType.INT16.value  : "h",
            ParamType.UINT32.value : "I",
            ParamType.INT32.value  : "i",
            ParamType.UINT64.value : "Q",
            ParamType.INT64.value  : "q",
            ParamType.FLT32.value  : "f",
            ParamType.FLT64.value  : "f",
            ParamType.DATE.value   : "",
            ParamType.STR.value    : "",
            ParamType.BLOB.value   : "",    
        }
        return specifiers[self.value]

    @property
    def underling_class(self):
        classes = {
            ParamType.BOOL.value   : bool,
            ParamType.UINT8.value  : int,
            ParamType.INT8.value   : int,
            ParamType.UINT16.value : int,
            ParamType.INT16.value  : int,
            ParamType.UINT32.value : int,
            ParamType.INT32.value  : int,
            ParamType.UINT64.value : int,
            ParamType.INT64.value  : int,
            ParamType.FLT32.value  : float,
            ParamType.FLT64.value  : float,
            ParamType.DATE.value   : datetime,
            ParamType.STR.value    : str,
            ParamType.BLOB.value   : bytearray,    
        }
        return classes[self.value]


    def get_size(self, capacity):
        classes = {
            ParamType.BOOL.value   : 1,
            ParamType.UINT8.value  : 1,
            ParamType.INT8.value   : 1,
            ParamType.UINT16.value : 2,
            ParamType.INT16.value  : 2,
            ParamType.UINT32.value : 4,
            ParamType.INT32.value  : 4,
            ParamType.UINT64.value : 8,
            ParamType.INT64.value  : 8,
            ParamType.FLT32.value  : 4,
            ParamType.FLT64.value  : 8,
            ParamType.DATE.value   : 8,
            ParamType.STR.value    : 1,
            ParamType.BLOB.value   : 1,    
        }
        return classes[self.value] * capacity



class Param:

    def __init__(self, parent: Sys, name:str, type: ParamType, capacity: int, value) -> 'Param':
        if(not isinstance(parent, Sys)):
            raise LibS3Error("Param's first argument must be of type Sys.")
        self._parent = parent
        self._name = Name(name)
        self._type  = type
        capacity = 1 if capacity == 0 else capacity
        self._size  = type.get_size(capacity)
        self._underling = type.underling_class
        self._be_encode_specifier = ">"+type.encode_specifier
        self._le_encode_specifier = "<"+type.encode_specifier
        self._value = value
        
        # Append to parent
        parent._Sys__add_param(self)

    @property
    def name(self) -> str:
        return self._name.as_str
    
    def get_name(self) -> Name:
        return self._name

    @property
    def index(self) -> int:
        return 0
        # TODO Implement

    @property
    def type(self)->str:
        return self._type.value

    @property
    def size(self)->str:
        return self._size

    @property
    def parent(self) -> Sys:
        return self._parent

    def get(self) -> Any:
        return self._value

    def set(self, value: Any):
        if(self._underling != type(value)):
            raise LibS3Error(f"Mismatching types, when trying to set Param '{self.name}' of type {self._underling}, with '{value}' of type {type(value)}.")
        self._value = value
            

    def path(self) -> Path:
        reverse = [self.name]
        parent = self.parent
        while parent is not None:
            reverse.append(parent.name)
            parent = parent.parent
        reverse.reverse()
        return Path(reverse)

    def to_chars(self) -> str:
        return self.__str__()


    def from_chars(self, value:str):
        try:
            self.set(self._underling(value))
        except:
            raise LibS3Error(f"Invalid data format when decoding param '{self.name}'.")


    def encode(self, to_big_endian:bool=False) -> bytes:
        if(to_big_endian):
            return struct.pack(self._be_encode_specifier, self._value)
        else:
            return struct.pack(self._le_encode_specifier, self._value)


    def decode(self, value:bytes, from_big_endian:bool=False):
        if from_big_endian:
            self._value, = struct.unpack( self._be_encode_specifier, value)
        else:
            self._value, = struct.unpack( self._le_encode_specifier, value)
        
        

    def __str__(self) -> str:
        return f"{self._value}"

    def __repr__(self) -> str:
        return self.__str__()
    
    def __format__(self, spec) -> str:
        return self.__str__().__format__(spec)

    def _asdict (self):
        return self._value





class ParamBool(Param):
    def __init__(self, parent: Sys, name: str, value=0) -> Param:
        super().__init__(parent, name, ParamType.BOOL, 1, value=value)


class ParamU8(Param):
    def __init__(self, parent: Sys, name: str, value=0) -> Param:
        super().__init__(parent, name, ParamType.UINT8, 1, value=value)


class ParamI8(Param):
    def __init__(self, parent: Sys, name: str, value=0) -> Param:
        super().__init__(parent, name, ParamType.INT8, 1,  value=value)


class ParamU16(Param):
    def __init__(self, parent: Sys, name: str, value=0) -> Param:
        super().__init__(parent, name, ParamType.UINT16, 1, value=value)


class ParamI16(Param):
    def __init__(self, parent: Sys, name: str, value=0) -> Param:
        super().__init__(parent, name, ParamType.INT16, 1,  value=value)


class ParamU32(Param):
    def __init__(self, parent: Sys, name: str, value=0) -> Param:
        super().__init__(parent, name, ParamType.UINT32, 1, value=value)


class ParamI32(Param):
    def __init__(self, parent: Sys, name: str, value=0) -> Param:
        super().__init__(parent, name, ParamType.INT32, 1, value=value)


class ParamU64(Param):
    def __init__(self, parent: Sys, name: str, value=0) -> Param:
        super().__init__(parent, name, ParamType.UINT64, 1, value=value)


class ParamI64(Param):
    def __init__(self, parent: Sys, name: str, value=0) -> Param:
        super().__init__(parent, name, ParamType.INT64, 1, value=value)


class ParamF32(Param):
    def __init__(self, parent: Sys, name: str, value=0) -> Param:
        super().__init__(parent, name, ParamType.FLT32, 1, value=value)


class ParamF64(Param):
    def __init__(self, parent: Sys, name: str, value=0) -> Param:
        super().__init__(parent, name, ParamType.FLT64, 1, value=value)


class ParamStr(Param):
    def __init__(self, parent: Sys, name: str, capacity:int, value="") -> Param:
        super().__init__(parent, name, ParamType.STR, capacity, value=value)

    def encode(self, to_big_endian:bool=False) -> bytes:
        return self._value.encode()

    def decode(self, value:bytes, from_big_endian:bool=False):
        self._value = value.decode()

class ParamBlob(Param):
    def __init__(self, parent: Sys, name: str, capacity:int, value=0) -> Param:
        super().__init__(parent, name, ParamType.BLOB, capacity, value=value)


    # TODO implement

    def encode(self, to_big_endian:bool=False) -> bytes:
        return self._value

    def decode(self, value:bytes, from_big_endian:bool=False):
        self._value = value



# Run with: python -m unittest -v libs3.param 
class TestParam(unittest.TestCase):

    def setUp(self):
        self.sys     = Sys(None, "sys")
        self.put_u8  = ParamU8(self.sys,  "u8",  1)
        self.put_i8  = ParamI8(self.sys,  "i8",  -1)
        self.put_u16 = ParamU16(self.sys, "u16", 123)
        self.put_i16 = ParamI16(self.sys, "i16", -123)
        self.put_u32 = ParamU32(self.sys, "u32", 123)
        self.put_i32 = ParamI32(self.sys, "i32", -123)
        self.put_u64 = ParamU64(self.sys, "u64", 123)
        self.put_i64 = ParamI64(self.sys, "i64", -123)
        self.put_f32 = ParamF32(self.sys, "f32", 4.20)
        self.put_f64 = ParamF64(self.sys, "f64", 4.20)
        self.put_str = ParamStr(self.sys, "str", 32, "Hi, Im libs3!")
        self.put_byt = ParamBlob(self.sys, "byt", 10, b'\x01\x02\x03\x04')
        self.subsys      = Sys(self.sys, "sub")
        self.sub_put_u8  = ParamU8(self.subsys,  "sub8",  1)

    def test_name(self):       
        self.assertEqual(self.put_u16.name, "u16")
        self.assertEqual(self.put_f32.name, "f32")
        self.assertEqual(self.put_str.name, "str")

    def test_wrong_formed(self):
        with self.assertRaises(LibS3Error):
            wrong_param = ParamU8("some", 1)
        with self.assertRaises(LibS3Error):
            wrong_param = ParamU8(self.sys, "", 1)

    def test_type(self):       
        self.assertEqual(self.put_u16.type, "u16")
        self.assertEqual(self.put_f32.type, "f32")
        self.assertEqual(self.put_str.type, "str")


    def test_set_get(self):
        self.put_u16.set(1010)
        self.put_f32.set(0.9898)
        self.put_str.set("Hi, you!!")

        self.assertEqual(self.put_u16.get(), 1010)
        self.assertEqual(self.put_f32.get(), 0.9898)
        self.assertEqual(self.put_str.get(), "Hi, you!!")


    def test_to_chars(self):
        self.put_u16.set(321)
        self.put_f32.set(0.42)
        self.put_str.set("Byee!!")

        self.assertEqual(self.put_u16.to_chars(), "321")
        self.assertEqual(self.put_f32.to_chars(), "0.42")
        self.assertEqual(self.put_str.to_chars(), "Byee!!")


    def test_from_chars(self):
        self.put_u16.from_chars("4321")
        self.put_f32.from_chars("0.4321")
        self.put_str.from_chars("Something else!!")

        self.assertEqual(self.put_u16.get(), 4321)
        self.assertEqual(self.put_f32.get(), 0.4321)
        self.assertEqual(self.put_str.get(), "Something else!!")


    def test_encode_little(self):
        self.put_u16.set(0x321)
        self.put_i16.set(-321) ## FEBF
        self.put_f32.set(1.23456)  # 1.23456 -> IEEE 754 -> 
        self.put_str.set("Byee!!")


        self.assertEqual(self.put_u16.encode(), b"\x21\x03")
        self.assertEqual(self.put_i16.encode(), b"\xbf\xfe")
        self.assertEqual(self.put_f32.encode(), b"\x10\x06\x9e\x3f")
        self.assertEqual(self.put_str.encode(), b"Byee!!")



    def test_encode_big(self):
        self.put_u16.set(0x321)
        self.put_i16.set(-321) ## FEBF
        self.put_f32.set(1.23456)  # 1.23456 -> IEEE 754 
        self.put_str.set("Byee!!")


        self.assertEqual(self.put_u16.encode(True), b"\x03\x21")
        self.assertEqual(self.put_i16.encode(True), b"\xfe\xbf")
        self.assertEqual(self.put_f32.encode(True), b"\x3f\x9e\x06\x10")
        self.assertEqual(self.put_str.encode(True), b"Byee!!")


    def test_decode_little(self): 
        self.put_u16.decode(b"\x94\x26") # = 9876
        self.put_i16.decode(b"\x6c\xd9") # = -9876
        self.put_f32.decode(b"\x19\x04\x1e\x41")  # = 9.876 411e0419
        self.put_str.decode(b"Hi again!!")


        self.assertEqual(self.put_u16.get(), 9876)
        self.assertEqual(self.put_i16.get(), -9876)
        self.assertAlmostEqual(self.put_f32.get(), 9.876, places=4)
        self.assertEqual(self.put_str.get(), "Hi again!!")



    def test_decode_big(self): 
        self.put_u16.decode(b"\x26\x94", True) # = 9876
        self.put_i16.decode(b"\xd9\x6c", True) # = -9876
        self.put_f32.decode(b"\x41\x1e\x04\x19", True)  # = 9.876 411e0419
        self.put_str.decode(b"Hi again!!", True)


        self.assertEqual(self.put_u16.get(), 9876)
        self.assertEqual(self.put_i16.get(), -9876)
        self.assertAlmostEqual(self.put_f32.get(), 9.876, places=4)
        self.assertEqual(self.put_str.get(), "Hi again!!")


    def test_decode_path(self): 
        path = self.put_u8.path()
        self.assertEqual(path.len, 2)
        self.assertEqual(path.__str__(), "sys:u8")

        path = self.sub_put_u8.path()
        self.assertEqual(path.len, 3)
        self.assertEqual(path.__str__(), "sys:sub:sub8")


    def test_mismatch_type(self):
        with self.assertRaises(LibS3Error):
            self.put_u16.set(1.23)
        with self.assertRaises(LibS3Error):
            self.put_f32.set(123)
        with self.assertRaises(LibS3Error):
            self.put_str.set(123)

if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python
__author__      = "Olman Quiros Jimenez"
__copyright__   = "Copyright 2024"
__license__     = "MIT"


import unittest
from typing import List, Optional, Any
from s3py import LibS3Error


class Path:

    _stack = []
    separator_char=":"

    def __init__(self, init_list=List) -> None:
        if init_list:
            self._stack = init_list


    # @classmethod
    # def from_str(self, str_path:str, separator_char=':'):
    #     if not isinstance(str_path, str):
    #         raise LibS3Error("Path().from_str() expects a string as input.")
    #     return Path(str_path.split(separator_char))


    def push(self, element:Any) -> None:
        self._stack.append(element)


    def pop(self)  -> Optional[Any] :
        if len(self._stack):
            return self._stack.pop()
        else:
            raise LibS3Error("Path is empty.")

    def peek(self)  -> Optional[Any] :
        if len(self._stack):
            return self._stack[len(self._stack)-1]
        else:
            raise LibS3Error("Path is empty.")


    def __getitem__(self, index:int)  -> Optional[Any] :
        if index < len(self._stack):
            return self._stack[index]
        else:
            raise LibS3Error("Index out of is Path range.")

    @property
    def len(self) -> int:
        return len(self._stack)


    def __str__(self) -> str:
        return self.separator_char.join([x.__str__() for x in self._stack])
    

    def __repr__(self) -> str:
        return self.__str__()


    def __iter__(self):
        self._iter_idx = 0 
        return self
    
    def __next__(self):
        try:
            result = self._stack[self._iter_idx]
        except IndexError:
            raise StopIteration
        self._iter_idx += 1
        return result





class NamePath(Path):
    
    def __init__(self, str_path=str, separator_char=':') -> None:
        if not isinstance(str_path, str):
            raise LibS3Error("NamePath() expects a string as input.")
        super().__init__(str_path.split(separator_char))



    
 # Run with: python -m unittest -v libs3.path 
class TestPath(unittest.TestCase):


    
    def test_path_from_str(self):
        path = NamePath("uno:dos:tres")
        self.assertEqual(path.len, 3)
        self.assertEqual(path.pop(), "tres")
        self.assertEqual(path.len, 2)
        self.assertEqual(path.pop(), "dos")
        self.assertEqual(path.len, 1)
        self.assertEqual(path.pop(), "uno")
        self.assertEqual(path.len, 0)

    def test_path_iter(self):
        path = NamePath("uno:dos:tres")
        idx = 0
        for i in path:
            self.assertEqual(i, path[idx])
            idx += 1


if __name__ == '__main__':

    unittest.main()


#!/usr/bin/env python
__author__      = "Olman Quiros Jimenez"
__copyright__   = "Copyright 2024"
__license__     = "MIT"


import unittest
from typing import  Union

from s3py import LibS3Error, Sys, Param, ParamType, NamePath, Group, _root, _group_list
from .s3_pprint import  pprint_param_line
from threading import Lock



class SingletonMeta(type):

    _instances = {}
    _lock: Lock = Lock()

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]


class ParamDB(metaclass=SingletonMeta):
    def __init__(self) -> None:
        pass

        
    def print_tree(self, index:bool = False, type:bool = False,  size:bool = False):
        _root.pprint(index, type, size)

    def search_sys(self, path:Union[str, NamePath]):
        if isinstance(path, str):
            path = NamePath(path)

        cur_sys = _root
        for sys_name in path:
            cur_sys = cur_sys.get_sys(sys_name)
        return cur_sys
    

    def search_param(self, path:Union[str, NamePath]):
        if isinstance(path, str):
            path = NamePath(path)
        param_name = path.pop()
        sys = self.search_sys(path)
        return sys.get_param(param_name)


    def search_sys_by_idx(self, index:int):
        # TODO implement
        pass


    def search_param_by_idx(self, index:int):
        # TODO implement
        pass

    def spawn_param(self, parent:Sys, name:str, type: ParamType, capacity:int=1, default:str=None):
        # Check if param with the same name already exists, to avoid repeated params
        try:
            result = parent.get_param(name)
        except:
            result =  Param(parent, name, type, capacity, default)
        result.from_chars(default)
        return result


    def spawn(self, path:NamePath, type: ParamType, capacity:int=1, default:str=None):
        if isinstance(path, str):
            path = NamePath(path)

        if path.len < 2:
            raise LibS3Error("Path for spawning a parm shall contain at leas 2 names.")
        # Extract name of param
        param_name = path.pop()        
        # Create all required paren systems in path
        cur_sys = _root
        for sys_name in path:
            try:
                cur_sys = cur_sys.get_sys(sys_name)
            except:
                cur_sys = Sys(cur_sys, sys_name)
        # Create the param
        return self.spawn_param(cur_sys, param_name, type, capacity, default)


    def __spawn_from_dict(self, sys_as_dict:dict, parent:Sys):
        try:
            cur_sys = parent.get_sys(sys_as_dict["sys"])
        except:
            cur_sys = Sys(parent, sys_as_dict["sys"])

        if "params" in sys_as_dict.keys():
            for param in sys_as_dict["params"]:
                capacity = param["size"] if "size"in param.keys() else 1
                default = param["default"] if "default"in param.keys() else None
                self.spawn_param(cur_sys, param["param"], ParamType(param["type"]), capacity, default)
                # Add param to corresponding groups
                if "groups" in param.keys():
                    for group_name in param["groups"]:
                        try:
                            group = self.get_group(group_name)
                        except:
                            group = Group(group_name)
                        group.add_param(param)
                        

        if "subsys" in sys_as_dict.keys():
            for sys in sys_as_dict["subsys"]:
                self._ParamDB__spawn_from_dict(sys, cur_sys)


    def spawn_from_dict(self, sys_as_dict:dict):
        self._ParamDB__spawn_from_dict(sys_as_dict, _root)


    def pprint_groups(self):
        for name, group in _group_list.items():
            print(f"{group.name}[{group.len}]")
    

    def get_group(self, name) -> Group:
        try:
            return _group_list[name]
        except:
            raise LibS3Error(f"Group '{name}' not found.")


# Run with: python -m unittest -v libs3.param_db 
class TestParamDB(unittest.TestCase):

    def setUp(self):
        self.db = ParamDB()

    def test_spawn(self):
        param_one =  self.db.spawn("sys:one", ParamType.UINT8, 1, "12")
        self.assertEqual(param_one, self.db.search_param("sys:one"))
        self.assertEqual(12, self.db.search_param("sys:one").get())
        self.assertEqual("u8", self.db.search_param("sys:one").type)
        self.assertEqual(1, self.db.search_param("sys:one").size)

        param_two =  self.db.spawn("sys:two", ParamType.FLT32, 1, "12.24")
        self.assertEqual(param_two, self.db.search_param("sys:two"))
        self.assertEqual(12.24, self.db.search_param("sys:two").get())
        self.assertEqual("f32", self.db.search_param("sys:two").type)
        self.assertEqual(4, self.db.search_param("sys:two").size)


        param_two =  self.db.spawn("sys:sub:thre", ParamType.STR, 10, "Wololo")
        self.assertEqual(param_two, self.db.search_param("sys:sub:thre"))
        self.assertEqual("Wololo", self.db.search_param("sys:sub:thre").get())
        self.assertEqual("str", self.db.search_param("sys:sub:thre").type)
        self.assertEqual(10, self.db.search_param("sys:sub:thre").size)


    def test_spawn_from_dict(self):


        

        self.db.spawn_from_dict({"sys": "dict", 
                                 "params": [{"param":"one", "type": "i32", "default": "123" }], 
                                 "subsys": [
                                     {"sys": "sub",
                                      "params":[{"param":"two", "type": "str", "size": 10, "default": "Hohoho" }]}
                                 ]})

        self.assertEqual(123, self.db.search_param("dict:one").get())
        self.assertEqual("i32", self.db.search_param("dict:one").type)
        self.assertEqual(4, self.db.search_param("dict:one").size)

        self.assertEqual("Hohoho", self.db.search_param("dict:sub:two").get())
        self.assertEqual("str", self.db.search_param("dict:sub:two").type)
        self.assertEqual(10, self.db.search_param("dict:sub:two").size)



if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python
__author__      = "Olman Quiros Jimenez"
__copyright__   = "Copyright 2024"
__license__     = "MIT"


import unittest
from s3py import LibS3Error

class Name:

    def __init__(self, new_name:str) -> 'Name':
        if(not new_name):
            raise LibS3Error("The Name of a Param/Sys can't be null initialized.")
        if(len(new_name) > 4):
            raise LibS3Error(f"Name '{new_name}' is too long, up to 4 chars allowed for a Name.")
        self._name = new_name
        chars = [b for b in new_name]  
        self._id = int.from_bytes(new_name.encode('utf-8'), "little") 

    @property
    def as_str(self) -> str:
        return self._name

    @property
    def as_int(self)-> int:
        return self._id

    def __str__(self) -> str:
        return f"{self._name}"

    def __repr__(self) -> str:
        return self.__str__()

    def __format__(self, spec) -> str:
        return self.__str__().__format__(spec)

    def _asdict (self):
        return self._name
    

    def __eq__(self, other:'Name') -> bool:

        if(isinstance(other, Name)):
            if(self.as_int == other.as_int):
                return True
        elif(isinstance(other, int)):
            if(self.as_int == other):
                return True
        elif(isinstance(other, str)):
            if(self.as_str == other):
                return True
        return False



# Run with: python -m unittest -v libs3.name 
class TestName(unittest.TestCase):
    pass

if __name__ == '__main__':
    unittest.main()


from pprint import pprint
from s3py import * #, ParamType #, Group


#### Create some Params and Systems ####
my_sys = Sys  (None, "my")
my_bol = ParamBool(my_sys, "bol", True)
my_int = ParamU32(my_sys, "int", 123)
my_flt = ParamF32(my_sys, "flt", 4.20)
my_str = ParamStr(my_sys, "str", "Hi, Im libs3!")


# Get the value of a Param
value = my_int.get()
print(f"Param {my_int.name} is {value}")

# Set the value of a Param
my_int.set(456)
print(f"Param {my_int.name} is {my_int.get()}")



#### Create a hierarchical structure of Sys and Params ####
adcs             = Sys      (None,        "adcs")
adcs_status      = ParamU8  (adcs,        "stat", 1)
adcs_error       = ParamBool(adcs,       "err",  False)
adcs_wheel1      = Sys     (adcs,        "whl1")
adcs_wheel1_volt = ParamF32(adcs_wheel1, "volt", 0.0)
adcs_wheel1_rpm  = ParamF32(adcs_wheel1, "rpm",  0.0)
adcs_wheel2      = Sys     (adcs,        "whl2")
adcs_wheel2_volt = ParamF32(adcs_wheel2, "volt", 0.0)
adcs_wheel2_rpm  = ParamF32(adcs_wheel2, "rpm",  0.0)
adcs_wheel3      = Sys     (adcs,         "whl3")
adcs_wheel3_volt = ParamF32(adcs_wheel3, "volt", 0.0)
adcs_wheel3_rpm  = ParamF32(adcs_wheel3, "rpm",  0.0)

adcs_power       = Sys     (adcs,         "pwr")
adcs_power_lim   = Sys     (adcs_power,   "lim")
adcs_power_lim_volt = ParamF32(adcs_power_lim, "volt", 10)
adcs_power_lim_cur  = ParamF32(adcs_power_lim, "cur",  2)


# Alternative access to the nested Params
print(f"ADCS status is: {adcs['stat']}")


# Also possible to set the value
adcs["whl2"]["volt"].set(8.2)
adcs["whl3"]["volt"].set(8.1)


# Print the whole collection
print("\nADCS Param structure:")
adcs.pprint()




#### Get values as dict ####
new_values = {"stat": 2, "whl1":{"volt": 5.0}, "whl2":{"volt": 6.0}, "whl3":{"volt": 7.0}}
adcs.set(new_values)


print("\nadcs sys as dict:")
pprint(adcs.to_dict())




#### Create a Group of Params ####
# Usefully to represent Domain-related groups of Params
wheel_volt = Group("volt",[ # Provide a name  
                adcs_wheel1_volt,
                adcs_wheel2_volt,
                adcs_wheel3_volt
                ])

# Print the whole collection
print("\nVolt of all wheels:")
wheel_volt.pprint()



# Every top-level system (system without a parent) is automatically 
# registered into a local data base: the ParamDB singleton class.
db = ParamDB()

print("\nPrint everything:")
db.print_tree()


print("\nAlso existing groups:")
db.pprint_groups()


# The ParamDB works as an interface to the outside. It provide means to query
# Params.
print(f"\nadcs:whl3:volt is {db.search_param(NamePath('adcs:whl3:volt')).get()}")

# ParamDB can create new Params and Systems at runtime based on a dictionary
# specification
db.spawn_from_dict({"sys": "dict", 
                                 "params": [{"param":"one", "type": "i32", "default": "123" }], 
                                 "subsys": [
                                     {"sys": "sub",
                                      "params":[{"param":"two", "type": "str", "size": 10, "default": "Hohoho" }]}
                                 ]})



# #### Serialization ####

# # Serialize a Collection
# print("\nCollection to a JSON string:")
# print(JsonParamEncoder().encode(adcs))

# # Deserialize a Collection
# print("\nLoad a JSON string to a Collection:")
# json_str = '{"stat": 3,  "whl1": {"volt": 8.0, "rpm": 100}, "whl2": {"volt": 8.2, "rpm": 95}, "whl3": {"volt": 8.1, "rpm": 90}}'
# JsonParamDecoder().decode(adcs, json_str)
# pprint(adcs)

# # Serialize a Group
# print("\n\nGroup to a JSON string:")
# print(JsonParamEncoder().encode(wheel_volt))

# # Deserialize a Group
# print("\nLoad a JSON string to a Group:")
# json_str = '{"whl1": 5.6, "whl2": 2.1, "whl3": 8.0}'
# JsonParamDecoder().decode(wheel_volt, json_str)
# pprint(wheel_volt)


#!/usr/bin/env python
__author__      = "Olman Quiros Jimenez"
__copyright__   = "Copyright 2024"
__license__     = "MIT"


import unittest
from typing import List, Dict, Optional, Union
from s3py import LibS3Error, Name
from .s3_pprint import pprint_sys_line, pprint_param_line

# Forward declaration
def _add_sys_to_root(sys): pass

class Sys():

    def __init__(self, parent:'Sys', name:str) -> 'Sys':
        self._parent = parent
        self._name = Name(name)
        if(parent):
            parent.__add_sys(self)
        else:
            _add_sys_to_root(self)  # Add top-levels sys to the root 
        self._subsys = []
        self._params = []
        

    @property
    def name(self) -> str:
        return self._name.as_str
    
    def get_name(self) -> Name:
        return self._name

    @property
    def index(self) -> int:
        return 0
        # TODO Implement

    @property
    def type(self) -> str:
        return "sys"
    
    @property
    def parent(self) -> Optional['Sys']:
        return self._parent

    def __add_sys(self, sys):
        self._subsys.append(sys)


    def __add_param(self, param):
        self._params.append(param)


    def get_sys(self, name)-> 'Sys':
        for s in self._subsys:
            if( s.name == name ):
                return s
        raise  LibS3Error(f"Sys '{name}' not found in '{self.name}'!")


    def get_param(self, name)-> 'Param':
        for p in self._params:
            if( p.name == name ):
                return p
        raise  LibS3Error(f"Param '{name}' not found in '{self.name}'!")
    

    def __getitem__(self, name) -> Union['Param', 'Sys'] :
        try:
            return self.get_param(name)
        except:
            pass
        try:
            return self.get_sys(name)
        except:
            raise  LibS3Error(f"Item '{name}' not found!")
    

    def to_dict(self) -> dict:
        result = {}
        for param in self._params:
            result[param.name] = param.get()
        for sys in self._subsys:
            result[sys.name] = sys.to_dict()
        return result
    

    def to_flat_dict(self)-> List[dict]:
        pass


    def set(self, data : dict):
        for key, value in data.items():
            self[key].set(value)
            


    def _pprint(self, depth, index, type, size): 
        for param in self._params:
            print(pprint_param_line(param, depth, index, type, size))

        for sys in self._subsys:
            print(pprint_sys_line(sys, depth, index, type, size))
            sys._pprint(depth+1, index, type,  size)



    def pprint(self, index:bool = False, type:bool = False,  size:bool = False): 
        self._pprint(0, index, type,  size)


# Root system
_root = Sys(None, "/")

def _add_sys_to_root(sys):
    if sys != _root:
        _root._Sys__add_sys(sys)




# Run with: python -m unittest -v libs3.sys 
class TestSys(unittest.TestCase):

    def setUp(self):
        self.parent = Sys(None, "prnt")

    def test_nested_name(self):
        print(self.parent)
        self.assertEqual(self.parent.name, "prnt")

if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python
__author__      = "Olman Quiros Jimenez"
__copyright__   = "Copyright 2024"
__license__     = "MIT"


from .protocol_message import RequestMessage, RequestCode



class CloneRequest(RequestMessage):
    '''
    Clone Request

    Get all Collections and Params from a Node
    '''
    def __init__(self) -> 'CloneRequest':
        super().__init__(RequestCode.CLONE)



class PullRequest(RequestMessage):
    '''
    Pull Request

    Synchronizes Collections of Remote Node given a commit hash.
    '''
    def __init__(self, commit_hash:str) -> 'CloneRequest':
        super().__init__(RequestCode.PULL)




class SubscribeRequest(RequestMessage):
    '''
    Subscribe Request

    Subscribes to a topic from a Remote Node.
    '''
    def __init__(self, topic:str) -> 'CloneRequest':
        super().__init__(RequestCode.SUBSCRIBE)



#!/usr/bin/env python
__author__      = "Olman Quiros Jimenez"
__copyright__   = "Copyright 2024"
__license__     = "MIT"


import unittest
from typing import List, Dict
from s3py import  Sys
from s3py.param import *
from .s3_pprint import  pprint_param_line


_group_list = dict()

class Group():

    def __init__(self, name:str, param_list: List[Param] = None) -> 'Group':
        self._name = name
        self._size = 0
        if param_list:
            self._list = param_list
            for param in self._list:
                if not isinstance(param, Param):
                    raise LibS3Error("Elements in Group initializer list must be subclass of Param.")
                self._size += param.size
            # Add to group list
            _group_list[name] = self
        else:
            self._list = []

    @property
    def name(self) -> str:
        return self._name
    
    @property
    def len(self) -> int:
        '''
        Number of elements in Group
        '''
        return len(self._list)

    @property
    def size(self) -> int:
        '''
        Sum of sizes of all Group's parameters
        '''
        return self._size


    def add_param(self, param:Param):
        if not isinstance(param, Param):
            raise LibS3Error("Elements to add to a Group must be subclass of Param.")
        if param not in self._list:
            self._list.append(Param)
            self._size += param.size


    def pprint(self, index:bool = False, type:bool = False,  size:bool = False): 
        for param in self._list:
            print(f"{param.path()}:\t{param.get()}")



    def encode(self, to_big_endian:bool=False) -> bytearray:
        result = bytearray(self.size)
        idx = 0
        
        for param in self._list:
            result[idx:idx+param.size] = param.encode(to_big_endian)
            idx += param.size
        return result


    def decode(self, value:bytes, from_big_endian:bool=False) -> int:
        result = 0
        idx = 0
        
        if len(value) < self.size:
            return result

        for param in self._list:
            param.decode(value[idx:idx+param.size], from_big_endian)
            result += 1
            idx += param.size
        return result





# Run with: python -m unittest -v libs3.group 
class TestParam(unittest.TestCase):
    def setUp(self):
        self.sys     = Sys(None, "sys")
        self.put_u16 = ParamI16(self.sys, "u16")
        self.put_i16 = ParamI16(self.sys, "i16")
        self.put_f32 = ParamF32(self.sys, "f32")
        self.put_str = ParamStr(self.sys, "str", 10)

        self.test_group = Group("test_group", [
            self.put_u16,
            self.put_i16,
            self.put_f32,
            self.put_str
        ])

    def test_size(self): 
        self.assertEqual(self.test_group.len, 4)
        self.assertEqual(self.test_group.size, 18)

    def test_decode_little(self): 

        self.put_u16.set(0)
        self.put_i16.set(0)
        self.put_f32.set(0.0)
        self.put_str.set("")

        data = b"\x94\x26\x6c\xd9\x19\x04\x1e\x41Hi again!!"

        self.assertEqual(self.test_group.decode(data), 4)

        self.assertEqual(self.put_u16.get(), 9876)
        self.assertEqual(self.put_i16.get(), -9876)
        self.assertAlmostEqual(self.put_f32.get(), 9.876, places=4)
        self.assertEqual(self.put_str.get(), "Hi again!!")




    def test_decode_big(self): 
        self.put_u16.set(0)
        self.put_i16.set(0)
        self.put_f32.set(0.0)
        self.put_str.set("")

        data = b"\x26\x94\xd9\x6c\x41\x1e\x04\x19Hi again!!"

        self.assertEqual(self.test_group.decode(data, True), 4)

        self.assertEqual(self.put_u16.get(), 9876)
        self.assertEqual(self.put_i16.get(), -9876)
        self.assertAlmostEqual(self.put_f32.get(), 9.876, places=4)
        self.assertEqual(self.put_str.get(), "Hi again!!")



    def test_encode_little(self): 

        self.put_u16.set(9876)
        self.put_i16.set(-9876)
        self.put_f32.set(9.876)
        self.put_str.set("Hi again!!")

        expected = b"\x94\x26\x6c\xd9\x19\x04\x1e\x41Hi again!!"

        self.assertEqual(self.test_group.encode(), expected)


    def test_encode_big(self): 

        self.put_u16.set(9876)
        self.put_i16.set(-9876)
        self.put_f32.set(9.876)
        self.put_str.set("Hi again!!")

        expected = b"\x26\x94\xd9\x6c\x41\x1e\x04\x19Hi again!!"

        self.assertEqual(self.test_group.encode(True), expected)

if __name__ == '__main__':
    unittest.main()